;; In the card game poker, a hand consists of five cards and are ranked, from lowest to highest, in the following way:

;;     * High Card: Highest value card.
;;     * One Pair: Two cards of the same value.
;;     * Two Pairs: Two different pairs.
;;     * Three of a Kind: Three cards of the same value.
;;     * Straight: All cards are consecutive values.
;;     * Flush: All cards of the same suit.
;;     * Full House: Three of a kind and a pair.
;;     * Four of a Kind: Four cards of the same value.
;;     * Straight Flush: All cards are consecutive values of same suit.
;;     * Royal Flush: Ten, Jack, Queen, King, Ace, in same suit.

;; The cards are valued in the order:
;; 2, 3, 4, 5, 6, 7, 8, 9, 10, Jack, Queen, King, Ace.

;; If two players have the same ranked hands then the rank made up of the highest value wins; for example, a pair of eights beats a pair of fives (see example 1 below). But if two ranks tie, for example, both players have a pair of queens, then highest cards in each hand are compared (see example 4 below); if the highest cards tie then the next highest cards are compared, and so on.

;; Consider the following four hands dealt to two players:
;; Hand Player 1 Player 2 Winner
;; 1 5H 5C 6S 7S KD
;; Pair of Fives
;; 2C 3S 8S 8D TD
;; Pair of Eights
;; Player 2
;; 2 5D 8C 9S JS AC
;; Highest card Ace
;; 2C 5C 7D 8S QH
;; Highest card Queen
;; Player 1
;; 3 2D 9C AS AH AC
;; Three Aces
;; 3D 6D 7D TD QD
;; Flush with Diamonds
;; Player 2
;; 4 4D 6S 9H QH QC
;; Pair of Queens
;; Highest card Nine
;; 3D 6D 7H QD QS
;; Pair of Queens
;; Highest card Seven
;; Player 1
;; 5 2H 2D 4C 4D 4S
;; Full House
;; With Three Fours
;; 3C 3D 3S 9S 9D
;; Full House
;; with Three Threes
;; Player 1

;; The file, poker.txt, contains one-thousand random hands dealt to two players. Each line of the file contains ten cards (separated by a single space): the first five are player one's cards and the last five are player two's cards. You can assume that all hands are valid (no invalid characters or repeated cards), each player's hand is in no specific order, and in each hand there is a clear winner.

;; How many hands does player one win?


(defvar *the-file* "/home/mortis/personal/projects/project-euler.net/data-files/poker.txt")

;; (asdf-install:install 'cl-utilities)
(require 'cl-utilities)


(defun parse-hands (line)
  (let ((cards (cl-utilities:split-sequence #\Space line)))
    (list
     (subseq cards 0 5)
     (subseq cards 5 10))))

(assert (equal
	 (parse-hands "9S 3H 9D JD 5C 4D 9H AS TC QH")
	 '(("9S" "3H" "9D" "JD" "5C") ("4D" "9H" "AS" "TC" "QH"))))


(defun get-hands (file)
  (with-open-file (in file :direction :input)
    (loop for line = (read-line in nil nil)
	  while line
	  collect (parse-hands line))))

(defvar *all-hands* (get-hands *the-file*))

;; (length *all-hands*)

(defun hand-matches (pat hand)
  (loop for card in pat
	do
	(when (not (member card hand :test #'equal))
	  (return-from hand-matches nil)))
  hand)

(assert (hand-matches '("TC") '("TC")))
(assert (not (hand-matches '("TC") '("9C"))))

(defun card-val->score (cval)
  (let ((val (subseq cval 0 1)))
    (cond
      ((equal "2" val) 2)
      ((equal "3" val) 3)
      ((equal "4" val) 4)
      ((equal "5" val) 5)
      ((equal "6" val) 6)
      ((equal "7" val) 7)
      ((equal "8" val) 8)
      ((equal "9" val) 9)
      ((equal "T" val) 10)
      ((equal "J" val) 11)
      ((equal "Q" val) 12)
      ((equal "K" val) 13)
      ((equal "A" val) 14))))


(defun card< (left right)
  (let ((lval  (subseq left 0 1))
	(rval  (subseq right 0 1)))
    (< (card-val->score lval)
       (card-val->score rval))))


(defun high-card (hand)
  (loop for card in hand
	with maxcard = (first hand)
	finally (return maxcard)
	do
	(when (card< maxcard card)
	  (setf maxcard card))))

(assert (equal "AD" (high-card '("AD" "9D"))))

(defun card-value (card)
  (subseq card 0 1))

(defun group-by-value (hand)
  (let ((groups (make-hash-table :test #'equal)))
    (loop 
     for card in hand
     for cval = (card-value card) then (card-value card)
     for set = (gethash cval groups) then (gethash cval groups)
     do
     (if (not set)
	 (setf (gethash cval groups) (list card))
	 (setf (gethash cval groups) (cons card (gethash cval groups)))))
    (loop for cval being the hash-keys of groups
	  collect (gethash cval groups))))

(defun one-pair? (hand)
  (let ((groups (group-by-value hand)))
    (and
     (= 4 (length groups))
     (= 1 (length (remove-if-not #'(lambda (group) (= 2 (length group))) groups)))
     (remove-if-not #'(lambda (group) (= 2 (length group))) groups))))

(assert (= 4 (length (group-by-value '("2H" "2D" "3S" "4C" "5H")))))
(assert (one-pair? '("2H" "2D" "3S" "4C" "5H")))
(assert (not (one-pair? '("2H" "2D" "2S" "4C" "5H"))))
(assert (not (one-pair? '("2H" "2D" "3S" "3C" "5H"))))
(assert (not (one-pair? '("2H" "3D" "4S" "5C" "6H"))))

(defun two-pair? (hand)
  (let ((counts (group-by-value hand)))
    (and
     (= 3 (length counts))
     (= 2 (length (remove-if-not #'(lambda (group) (= 2 (length group))) counts)))
     (remove-if-not #'(lambda (group) (= 2 (length group))) counts))))

(assert (two-pair? '("2H" "2D" "3S" "3C" "5H"))
(assert (not (two-pair? '("2H" "2D" "3S" "3C" "2H"))))
(assert (not (two-pair? '("5H" "2D" "6S" "3C" "2H"))))

(defun three-of-a-kind? (hand)
  (let ((groups (group-by-value hand)))
    (and
     (= 3 (length groups))
     (= 1 (length (remove-if-not #'(lambda (group) (= 3 (length group))) groups)))
     (remove-if-not #'(lambda (group) (= 3 (length group))) groups))))


(assert (not (three-of-a-kind? '("2H" "2D" "3S" "3C" "5H"))))
(assert (not (three-of-a-kind? '("2H" "2D" "3S" "3C" "3H"))))
(assert (three-of-a-kind? '("2H" "4D" "3S" "3C" "3H")))

(defun royal-flush? (hand)
  (or
   (hand-matches '("TH" "JH" "QH" "KH" "AH") hand)
   (hand-matches '("TD" "JD" "QD" "KD" "AD") hand)
   (hand-matches '("TS" "JS" "QS" "KS" "AS") hand)
   (hand-matches '("TC" "JC" "QC" "KC" "AC") hand)))

(assert (royal-flush? '("JC" "TC" "QC" "KC" "AC")))


(defun card-suit (card)
  (subseq card 1 2))

(defun get-suits (hand)
  (loop for card in hand
	collect
	(card-suit card)))

(defun get-values (hand)
  (loop for card in hand
	collect
	(card-value card)))

;; (get-values '("TC" "JC"))

(defun list-unique (things)
  (let ((ht (make-hash-table :test #'equal))
	(result (list)))
    (loop for thing in things
	  do
	  (when (not (gethash thing ht))
	    (setf (gethash thing ht) t)
	    (push thing result)))
    (reverse result)))

(defun list-unique-count (things)
  (let ((ht (make-hash-table :test #'equal)))
    (loop for thing in things
	  do
	  (if (not (gethash thing ht))
	      (setf (gethash thing ht) 1)
	      (setf (gethash thing ht) (1+ (gethash thing ht)))))
    (loop for key being the hash-keys of ht
	  collect (list key (gethash key ht)))))

;; (list-unique-count '(1 2 3 2 1 1))

(defun same-suits? (hand)
  (= 1 (length (list-unique (get-suits hand)))))

(assert (same-suits? '("TC" "JC" "QC" "KC" "AC")))
(assert (not (same-suits? '("TC" "JC" "QC" "KC" "AD"))))

(defun flush? (hand)
  (same-suits? hand))

(assert (flush? '("TC" "JC" "QC" "KC" "AC")))
(assert (not (flush? '("TC" "JC" "QC" "KC" "AD"))))

;;(sort '("8C" "TS" "KC" "9H" "4S") #'card<)

;; (first *all-hands*) (("8C" "TS" "KC" "9H" "4S") ("7D" "2S" "5D" "3S" "AC"))

(defun consecutive-cards? (hand)
  (let ((vals (sort (get-values hand) #'card<)))
    (loop for curr = (car vals) then (car rst)
	  for rst = (rest vals) then (rest rst)
	  for next = (car rst) then (car rst)
	  while rst
	  do
	  (unless (= 1 (- (card-val->score next)
			  (card-val->score curr)))
	    (return-from consecutive-cards? nil)))
    t))

(assert (not (consecutive-cards? '("8C" "TS" "KC" "9H" "4S"))))
(assert (consecutive-cards? '("8C" "9S" "TC" "JH" "QS")))

;(group-by-value '("2H" "2D" "3S" "4C" "5H"))

(defun straight-flush? (hand)
  (and
   (flush? hand)
   (consecutive-cards? hand)
   hand))

(assert (straight-flush? '("9C" "TC" "JC" "QC" "KC")))

(defun four-of-a-kind? (hand)
  (let ((counts (list-unique-count (get-values hand))))
    (and
     (= 2 (length counts))
     (= 1 (length (remove-if #'(lambda (pair) (= 4 (second pair))) counts))))))

(assert      (four-of-a-kind? '("2H" "2D" "2S" "2C" "3H")))
(assert (not (four-of-a-kind? '("2H" "2D" "2S" "3C" "3H"))))


(defun full-house? (hand)
  (let ((counts (list-unique-count (get-values hand))))
    (and
     (= 2 (length counts))
     (= 1 (length (remove-if #'(lambda (pair) (= 3 (second pair))) counts)))
     hand)))

(assert (full-house? '("3C" "3D" "3H" "2C" "2D")))
(assert (not (full-house? '("3C" "3D" "3H" "2C" "4D")))


(defun flush? (hand)
  (same-suits? hand))

(assert (not (flush? '("TC" "JC" "QC" "KC" "AD"))))
(assert (flush? '("TC" "JC" "QC" "KC" "AC")))

(defun straight? (hand)
  (and
   (consecutive-cards? hand)
   hand))

(assert (not(straight? '("9C" "JC" "QC" "KC" "AD"))))
(assert (straight? '("TC" "JC" "QC" "KC" "AD")))

(defun break-tie (p1 p2)
  (let ((c1 (first (reverse (sort (mapcar #'card-val->score p1) #'card<))))
	(c2 (first (reverse (sort (mapcar #'card-val->score p2) #'card<)))))
    (cond ((equal c1 c2)
	   'tie)
	  ((string< c1 c2)
	   'p1)
	  (t
	   'p2))))

(defun first-card-of-set (hand cnt)
  (let ((groups (group-by-value hand)))
    (loop for group in groups
	  do
	  (when (= cnt (length group))
	    (return-from first-card-of-set (first group))))
    nil))

;; (first-card-of-set '("2H" "2D" "2S" "2C" "3H") 4)


(defun remove-sets (hand &rest counts)
  (let ((groups (group-by-value hand)))
    (apply #'append (remove-if #'(lambda (group) 
				   (remove-if-not #'(lambda (cnt) (= cnt (length group)))
						  counts))
			       groups))))

;; (remove-sets '("2H" "2D" "2S" "2C" "3H") 4 1)

;;(defmacro dbg (&rest body) '())
(defmacro dbg (&rest body) `(progn ,@body))

(defun classify-hand (hand)
  "Classifys the poker hand.  Returns a list of the classification, a
ranked set of tie breakers and the hand.

  (classify-hand '(\"2H\" \"2D\" \"2S\" \"2C\" \"3H\")
    => (four-of-a-kind (\"3H\") (\"2H\" \"2D\" \"2S\" \"2C\" \"3H\"))

"
  (cond ((royal-flush? hand)
	 (list 'royal-flush () hand))
	((straight-flush? hand)
	 (list 'straight-flush (list (high-card hand)) hand))
	((four-of-a-kind? hand)
	 (list 'four-of-a-kind (list (high-card (remove-sets hand 1))
				      (high-card (remove-sets hand 4)))
	       hand))
	((full-house? hand)
	 (list 'full-house (list (high-card (remove-sets hand 2))
				 (high-card (remove-sets hand 3)))
	       hand))
	((flush? hand)
	 (list 'flush (list (high-card hand))
	       hand))
	((straight? hand)
	 (list 'straight (list (high-card hand)) hand))
	((three-of-a-kind? hand)
	 (list 'three-of-a-kind (list (high-card (remove-sets hand 1))
				      (high-card (remove-sets hand 3)))
	       hand))
	((two-pair? hand)
	 (list 'two-pair (list (high-card (first  (group-by-value (remove-sets hand 1))))
			       (high-card (second (group-by-value (remove-sets hand 1))))
			       (high-card (remove-sets hand 2)))
	       hand))
	((one-pair? hand)
	 (list 'one-pair (list (high-card (remove-sets hand 1))
			       (high-card (remove-sets hand 2)))
	       hand))
	(t
	 (list 'high-card (list (high-card hand)) hand))))

;; (classify-hand '("JC" "TC" "QC" "KC" "AC"))
;; (classify-hand '("JC" "TC" "QC" "KC" "9C"))
;; (classify-hand '("JC" "JS" "JH" "JD" "9C"))
;; (classify-hand '("3C" "3S" "3H" "2D" "2C"))
;; (classify-hand '("3C" "4C" "5C" "6C" "8C"))
;; (classify-hand '("3C" "4C" "5C" "6C" "7D"))
;; (classify-hand '("3C" "3S" "3D" "6C" "7D"))
;; (classify-hand '("3C" "3S" "2D" "2C" "7D"))
;; (classify-hand '("3C" "3S" "4D" "5C" "7D"))
;; (classify-hand '("3C" "8S" "4D" "5C" "7D"))

;; if I was going to do this again, I'd write a classifier to identify
;; each hand type, then use a comparison table determine the winner,
;; breaking ties as necessary, but the hand classifier would have made
;; this a lot easier to debug and work with...

(defun classified-hand-type->value (classified-hand)
  (case (car classified-hand)
    ('royal-flush 10)
    ('straight-flush 9)
    ('four-of-a-kind 8)
    ('full-house 7)
    ('flush 6)
    ('straight 5)
    ('three-of-a-kind 4)
    ('two-pair 3)
    ('one-pair 2)
    ('high-card 1)))

(defun who-wins? (p1 p2)
  (let* ((left   (classify-hand p1))
	 (lscore (classified-hand-type->value left))
	 (right  (classify-hand p2))
	 (rscore (classified-hand-type->value right)))
    (format t "p1: (~a)~a vs p1: (~a)~a~&" lscore left rscore right)
    (cond ((> lscore rscore)
	   'p1)
	  ((< lscore rscore)
	   'p2)
	  ((= lscore rscore)
	   (loop for lbreak in (second left)
		 for rbreak in (second right)
		 do
		 (when (card< lbreak rbreak)
		   (return-from who-wins? 'p2))
		 (when (card< rbreak lbreak)
		   (return-from who-wins? 'p1)))
	   'tie)
	  (t
	   (error "don't know how to find winner?")))))


;; need to score particular ties appropraitely...(8's win vs 5's)
(who-wins? '("5H" "5C" "6S" "7S" "KD") '("2C" "3S" "8S" "8D" "TD"))
(who-wins? '("5D" "8C" "9S" "JS" "AC") '("2C" "5C" "7D" "8S" "QH"))



(defun analyze-all-hands ()
  (let ((tot 0))
    (loop for pair in *all-hands*
	  for p1 = (first pair) then (first pair)
	  for p2 = (second pair) then (second pair)
	  do
	  (when (equal 'p1 (who-wins? p1 p2))
	    (incf tot)))
    (format t "tot:~a~&" tot)))

(time (analyze-all-hands))